from typing import Any, Protocol, TypeAlias, TypeGuard

from holm.fastapi import FastAPIDependency

Page: TypeAlias = FastAPIDependency[Any]
"""FastAPI dependency that returns the properties for the layout that wraps it
(for example a `Component`).

If the page is not wrapped by a layout, then it must return a `Component`
"""


class PageDefinition(Protocol):
    """Protocol definition for objects (usually modules) that define a page."""

    @property
    def page(self) -> Page: ...


def is_page_definition(obj: Any) -> TypeGuard[PageDefinition]:
    """Type guard for `PageDefinition`."""
    page = getattr(obj, "page", None)
    return callable(page)



__version__ = "0.1.0"

from .app import App as App
from .fastapi import FastAPIDependency as FastAPIDependency
from .fastapi import FastAPIErrorHandler as FastAPIErrorHandler
from .modules._error import ErrorHandlerMapping as ErrorHandlerMapping
from .modules._metadata import Metadata as Metadata



from collections.abc import Callable
from typing import Any, Protocol, TypeAlias, TypeGuard

from fastapi import APIRouter
from fasthx.htmy import HTMY

PlainAPIFactory: TypeAlias = Callable[[], APIRouter]
RenderingAPIFactory: TypeAlias = Callable[[HTMY], APIRouter]

APIFactory: TypeAlias = PlainAPIFactory | RenderingAPIFactory
"""
`APIRouter` factory definition.

It is a callable that optionally accepts an `HTMY` instance as an argument to make sure
the same renderer is used throughout the application if the API does HTML rendering.
Otherwise the argument can be omitted.
"""


class APIDefinition(Protocol):
    """Protocol definition for objects (usually modules) that define an API."""

    @property
    def api(self) -> APIFactory | APIRouter:
        """
        The property may be an `APIFactory` or a plain `APIRouter` instance."""
        ...


def is_api_definition(obj: Any) -> TypeGuard[APIDefinition]:
    """Type guard for `APIDefinition`."""
    api = getattr(obj, "api", None)
    # Check APIRouter first, because that's also callable.
    return isinstance(api, APIRouter) or callable(api)



from collections.abc import ItemsView, Iterator, KeysView, Mapping, ValuesView
from typing import Any, Protocol, TypeAlias, overload

from htmy import Component, ContextAware, as_component_sequence

from holm.fastapi import FastAPIDependency

MetadataMapping: TypeAlias = Mapping[Any, Any]
"""
A mapping representation of information to be placed in the `<head>` element of an HTML document.

It is usually a dictionary with string keys and values, but it is not enforced in runtime,
so the typing is more lenient.
"""


class MetadataOwner(Protocol):
    """
    Protocol definition for objects (usually modules) that define page metadata.
    """

    @property
    def metadata(self) -> MetadataMapping | FastAPIDependency[MetadataMapping | None]: ...


def get_metadata_dependency(obj: Any) -> FastAPIDependency[MetadataMapping | None]:
    """
    Loads and returns the FastAPI `MetadataMapping` dependency based on the metadata definition
    in the given object.

    If the object is not a `MetadataOwner`, the dependency will simply return `None`.
    """
    metadata = getattr(obj, "metadata", None)
    if metadata is None or isinstance(metadata, Mapping):
        return lambda: metadata

    if callable(metadata):
        return metadata  # type: ignore[no-any-return]

    raise ValueError(f"Invalid metadata object: {metadata}")


class Metadata(ContextAware):
    """
    Context-aware metadata provider.

    The class implements the `Mapping` protocol as a means of giving convenient access to
    the metadata it contains.
    """

    __slots__ = ("_metadata",)

    def __init__(self, metadata: MetadataMapping) -> None:
        """
        Initialization.

        Arguments:
            metadata: The actual metadata mapping.
        """
        self._metadata = metadata
        """The metadata mapping."""

    @overload
    def get(self, key: Any, default: None = None) -> Any | None: ...

    @overload
    def get(self, key: Any, default: Any) -> Any: ...

    def get(self, key: Any, default: Any | None = None) -> Any | None:
        """Implements `Mapping.get()`."""
        return self._metadata.get(key, default)

    def __contains__(self, key: Any) -> bool:
        """Implements the `Container` protocol."""
        return key in self._metadata

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Metadata) and self._metadata == other._metadata

    def __getitem__(self, key: Any) -> Any:
        """Implements the `Sequence` protocol."""
        return self._metadata[key]

    def __iter__(self) -> Iterator[Any]:
        """Implements the `Iterable` protocol."""
        return iter(self._metadata)

    def __len__(self) -> int:
        """Implements the `Sized` protocol."""
        return len(self._metadata)

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def items(self) -> ItemsView[Any, Any]:
        """Implements `Mapping.items()`."""
        return self._metadata.items()

    def keys(self) -> KeysView[Any]:
        """Implements `Mapping.keys()`."""
        return self._metadata.keys()

    def values(self) -> ValuesView[Any]:
        """Implements `Mapping.values()`."""
        return self._metadata.values()


def components_with_metadata(data: tuple[Component, MetadataMapping | None]) -> Component:
    """
    Function that converts a `Component` and optional `MetadataMapping` tuple into a `Component`
    in which the received component is wrapped in a `MetadataContext`, making the given metadata
    available in all wrapped components using `Metadata.from_context()`.
    """
    components, metadata = data
    # Wrap the rendered components in Metadata context provider component.
    # This allows `Metadata.from_context()` to be used in all wrapped components.
    # The received components object is usually a ComponentSequence (doctype and html components),
    # so converting to a component sequence seems a tiny bit more efficient.
    return (
        components
        if metadata is None
        else Metadata(metadata).in_context(*as_component_sequence(components))
    )



import inspect
from collections.abc import Awaitable, Iterable
from typing import cast

from fastapi import APIRouter, Depends, FastAPI
from fastapi.responses import HTMLResponse
from fasthx.htmy import HTMY
from htmy import Component, as_component_type

from ._model import AppConfig, AppNode, PackageInfo
from .fastapi import FastAPIDependency
from .modules._api import PlainAPIFactory, RenderingAPIFactory, is_api_definition
from .modules._error import is_error_handler_owner, register_error_handlers
from .modules._layout import (
    LayoutFactory,
    combine_layouts_to_dependency,
    empty_layout_dependency,
    is_layout_definition,
)
from .modules._metadata import MetadataMapping, components_with_metadata, get_metadata_dependency
from .modules._page import is_page_definition
from .typing import module_names


def App(*, app: FastAPI | None = None, htmy: HTMY | None = None) -> FastAPI:
    """
    Creates a FastAPI application with all the routes that are defined in the application package.

    Arguments:
        app: Optional FastAPI application to use. If `None`, a new instance will be created.
        htmy: Optional `fasthx.htmy.HTMY` instance to use for server-side rendering. If `None`,
            a default instance will be created.
    """
    if app is None:
        app = FastAPI()

    if htmy is None:
        htmy = HTMY()

    config = AppConfig.default()
    packages = _discover_app_packages(config)
    root_node = _build_app_tree(packages)

    # Register error handlers
    if pkg := root_node.package:
        register_error_handlers(app, pkg.import_module("error", is_error_handler_owner), htmy=htmy)

    # Build the API
    app.include_router(_build_api(root_node, htmy=htmy))

    return app


def _build_api(
    node: AppNode,
    *,
    base_layout_dep: FastAPIDependency[LayoutFactory] = empty_layout_dependency,
    htmy: HTMY,
) -> APIRouter:
    """
    Recursively builds an `APIRouter` based on the application defined by `node`.

    Arguments:
        node: Application definition.
        base_layout: The base layout dependency to use for the API.
        htmy: The `fasthx.htmy.HTMY` instance to use for server-side rendering.
    """
    layout_dep = base_layout_dep  # In case there is no package or it has no layout module.
    pkg = node.package
    api = _make_api_router_for_package(pkg, htmy)
    if pkg is not None:
        layout_module = pkg.import_module("layout", is_layout_definition)
        page_module = pkg.import_module("page", is_page_definition)

        layout_dep = combine_layouts_to_dependency(
            base_layout_dep, None if layout_module is None else layout_module.layout
        )
        page_dep = None if page_module is None else page_module.page

        if page_dep is not None:
            metadata_dep = get_metadata_dependency(page_module)

            async def path_operation(
                layout: LayoutFactory = Depends(layout_dep),  # noqa: B008
                metadata: MetadataMapping | None = Depends(metadata_dep),  # noqa: B008
                page: Component = Depends(page_dep),  # noqa: B008
            ) -> tuple[Component, MetadataMapping | None]:
                result = layout(as_component_type(page))
                # We must await here if result is an Awaitable, otherwise we would pass an
                # awaitable to htmy.page() and rendering that would fail.
                if isinstance(result, Awaitable):
                    result = await result

                return result, metadata

            # Register the route with rendering.
            api.get(
                "/",
                response_class=HTMLResponse,
                response_model=None,
                name="page",
                description=page_dep.__doc__,
                tags=["Page"],
            )(htmy.page(components_with_metadata)(path_operation))

    for sub_url, child_node in node.subtree.items():
        api.include_router(
            _build_api(child_node, base_layout_dep=layout_dep, htmy=htmy),
            prefix=sub_url,
        )

    return api


def _build_app_tree(packages: Iterable[PackageInfo]) -> AppNode:
    """
    Returns an `AppNode` that represents the entire package tree of the application.

    Arguments:
        packages: The packages to include in the tree.
    """
    root = AppNode("/")

    for p in packages:
        root.add(p)

    return root


def _discover_app_packages(config: AppConfig) -> set[PackageInfo]:
    """
    Discovers all packages that are part of the application and returns them as a set.
    """
    return {
        PackageInfo.from_marker_file(f, config=config)
        for f in config.app_dir.rglob("*.py")
        if f.stem in module_names
    }


def _make_api_router_for_package(pkg: PackageInfo | None, htmy: HTMY) -> APIRouter:
    """
    Creates an `APIRouter` for the given package.
    """
    api_module = None if pkg is None else pkg.import_module("api", is_api_definition)
    if api_module is None:
        return APIRouter()

    # api is either an APIRouter or a callable (is_api_definition validates that).
    api = api_module.api
    if isinstance(api, APIRouter):
        return api

    # Inspect the api callable instead of catching errors. Using a try-except with
    # a TypeError handler would hide many potential issues that would then be
    # hard to figure out without the exception trace.
    num_params = len(inspect.signature(api).parameters)
    if num_params == 0:
        api = cast(PlainAPIFactory, api)()
    elif num_params == 1:
        api = cast(RenderingAPIFactory, api)(htmy)

    if isinstance(api, APIRouter):
        return api

    raise ValueError(f"The api function of {cast(PackageInfo, pkg).package_name} must return an APIRouter.")



from collections.abc import Callable, Mapping
from typing import Any, Protocol, TypeAlias, TypeGuard

from fastapi import FastAPI, Request, Response
from fastapi.responses import HTMLResponse
from fasthx.htmy import HTMY

from holm.fastapi import FastAPIErrorHandler

ErrorHandlerMapping: TypeAlias = Mapping[type[Exception] | int, FastAPIErrorHandler]
"""
Mapping whose keys are exception types or HTTP status codes, and the corresponding
values are FastAPI error handlers.
"""


class ErrorHandlerOwner(Protocol):
    """
    Protocol definition for objects (usually modules) that define FastAPI exception handlers.
    """

    @property
    def handlers(self) -> ErrorHandlerMapping | Callable[[HTMY], ErrorHandlerMapping]: ...


def is_error_handler_owner(obj: Any) -> TypeGuard[ErrorHandlerOwner]:
    """Type guard for `ErrorHandlerOwner`."""
    handlers = getattr(obj, "handlers", None)
    return handlers is not None and (isinstance(handlers, Mapping) or callable(handlers))


def register_error_handlers(app: FastAPI, owner: ErrorHandlerOwner | None, *, htmy: HTMY) -> None:
    """
    Registers the error handlers defined by the given owner on the application.

    Arguments:
        app: The FastAPI application instance.
        owner: The error handler owner, if any.
        htmy: The HTMY instance.
    """
    if owner is None:
        return

    handlers = owner.handlers
    handler_mapping: ErrorHandlerMapping = handlers(htmy) if callable(handlers) else handlers
    for key, handler in handler_mapping.items():
        app.exception_handler(key)(wrap_error_handler(handler, htmy))


def wrap_error_handler(error_handler: FastAPIErrorHandler, htmy: HTMY) -> FastAPIErrorHandler:
    """
    Wraps the given error handler in a function that automatically renders `htmy` `Component` return values.
    """

    async def rendering_error_handler_wrapper(request: Request, error: Exception) -> Response:
        result = await error_handler(request, error)
        if isinstance(result, Response):
            return result

        try:
            return HTMLResponse(await htmy.render_component(result, request))
        except Exception as e:
            raise ValueError(
                "HTML rendering failed. Make sure your error handler returns "
                "either a FastAPI.Response or an htmy.Component!"
            ) from e

    return rendering_error_handler_wrapper



from typing import Literal, TypeAlias, get_args

PackageModuleName: TypeAlias = Literal["api", "layout", "page"]
"""Recognized module names."""

RootOnlyModuleName: TypeAlias = Literal["error"]
"""Modules names that are only supported at the root level."""

ModuleName: TypeAlias = PackageModuleName | RootOnlyModuleName
"""All recognized module names, including ones that are only supported at the root level."""

# get_args() doesn't return the correct tuple for the ModuleName union type.
module_names: set[ModuleName] = {*get_args(PackageModuleName), *get_args(RootOnlyModuleName)}
"""Set of all recognized module names."""

URLSegment: TypeAlias = str
"""
A URL that consists of a leading slash followed by a non-empty URL part
that includes no more slashes.

Valid examples: "/about".

Invalid examples: "/", "/about/us".
"""
